/**
 * @ingroup   browse68_devel
 * @file      SC68browse.c
 * @author    Ben(jamin) Gerard <ben@sashipa.com>
 * @date      1999/02/22
 * @brief     sc68 Browse Info Data Base management
 * @version   $Id$
 */

/*
 *                       sc68 - Browse info generator
 *         Copyright (C) 2001 Ben(jamin) Gerard <ben@sashipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

/* TODO :
 * + Add a selected field in browse entry
 * + Add an interface to (un)select entry with criteria
 *   e.g select(action, condition)
 *       action := [ clear set inverse ]
 *       condition := (($$name[||$$composer]==*Jess*[&&truc]))&&$$time<5:0
 *
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>

#include "browse68/SC68browse.h"
#include "file68/SC68error.h"
#include "file68/SC68file.h"
#include "browse68/SC68html.h"
#include "file68/SC68utils.h"
#include "file68/SC68os_general.h"

static SC68browse_info_t bi;
static int browse_read(FILE *f);

/* typedef for human readable capabilities */
typedef void (*  SC68pdebug_f) (const char *, ...);

/* Dummy debug function, does really nothing
 */
static void no_pdebug(const char *fmt, ...)
{
}

/* Current debug function
 */
static SC68pdebug_f pdebug = no_pdebug;

/* Set new function for debug info display.
 * This must function is printf like.
 * return previous function
 */
SC68pdebug_f SC68browse_set_pdebug(const SC68pdebug_f new_pdebug)
{
  SC68pdebug_f old = pdebug;

  pdebug = new_pdebug ? new_pdebug : no_pdebug;
  return old;
}

static int namecmp(char *a, char *b)
{
  return SC68utils_strcmp(a, b);
}

/* Get current browse info
 */
SC68browse_info_t *SC68browse_getcurrent(void)
{
  return &bi;
}

/* First time init : Clean struct & load Browse info if fname != NULL
 */
int SC68browse_init(char *fname)
{
  pdebug("SC68browse_init(%s)\n", fname ? fname : "<null>");
  memset(&bi, 0, sizeof(bi));
  /* Read default browse */
  if (fname != NULL)
    SC68browse_read(fname);

  SC68browse_sort(SC68browse_sort_by_author);

  return 0;
}

/* First time init (added for external utils) :
 * Clean struct & load Browse info from opened file
 */
int SC68browse_init_bis(FILE * f)
{
  int err = 0;

  memset(&bi, 0, sizeof(bi));
  /* Read default browse */
  if (f)
    browse_read(f);
  if (!err)
    SC68browse_sort(SC68browse_sort_by_author);
  return err;
}

/* Clean exit caller : free memory, clean struct
 */
int SC68browse_kill(void)
{
  SC68browse_file_t *be, *next;

  for (be = bi.head; be != NULL; be = next) {
    next = be->nxt;
    SC68_free(be);
  }
  memset(&bi, 0, sizeof(bi));
  return 0;
}


/* Alloc & fill a browse info entry struct
 */
static SC68browse_file_t *new_entry(SC68browse_file_t * model)
{
  SC68browse_file_t *be;
  int sz, addr, al, cl, ml, pl;

  /* Process string length */
  al = strlen(model->aname) + 1;
  cl = (model->cname != model->aname) ? strlen(model->cname) + 1 : 0;
  ml = strlen(model->mname) + 1;
  pl = strlen(model->path) + 1;
  sz = al + cl + ml + pl + sizeof(SC68browse_file_t);
  /* Memory alloc */
  if (be = (SC68browse_file_t *) SC68_malloc(sz), be == NULL) {
    SC68error_add("SC68browse_file_t new_entry : Can't alloc %d bytes", sz);
    return NULL;
  }
  /* Copy strings */
  memset(be, 0, sz);
  addr = (int) (be + 1);
  be->aname = strcpy((char *) addr, model->aname);
  addr += al;
  if (model->cname != model->aname) {
    be->cname = strcpy((char *) addr, model->cname);
    addr += cl;
  } else
    be->cname = be->aname;
  be->mname = strcpy((char *) addr, model->mname);
  addr += ml;
  be->path = strcpy((char *) addr, model->path);
  addr += pl;
  /* Copy others */
  be->time = model->time;
  be->flags = model->flags;
  be->ntrack = model->ntrack;

  return be;
}



/* Set bit for different fields
 * 0  : music name
 * 1  : author
 * 2  : composer
 * 3  : path
 * 4 : time
 * 5 : flags
 * 6 : ntracks
 */
static int browse_cmp(const SC68browse_file_t *b1, const SC68browse_file_t *b2)
{
  int r = 0;
  char *c1, *c2;
  c1 = ( b1->cname==NULL )? b1->aname : b1->cname;
  c2 = ( b2->cname==NULL )? b2->aname : b2->cname;
  r |= (b1->mname!=b2->mname && !!namecmp(b1->mname,b2->mname)) << 0;
  r |= (b1->aname!=b2->aname && !!namecmp(b1->aname,b2->aname)) << 1;
  r |= (c1!=c2               && !!namecmp(c1,c2)              ) << 2;
  r |= (!!namecmp(b1->path,b2->path)) << 3;
  r |= (b1->time   != b2->time)   << 4;
  r |= (b1->flags  != b2->flags)  << 5;
  r |= (b1->ntrack != b2->ntrack) << 6;
  return r;
}

/* Find same path name OR (same all but path)
 * in Browse-Info-Data-Base
 */
SC68browse_file_t *SC68browse_exist(SC68browse_file_t *model)
{
  SC68browse_file_t *be;
  if(model==NULL) return NULL;
  for(be=bi.head; be!=NULL; be=be->nxt)
  {
    int r;
    r = browse_cmp(model,be);
    /* If same path or only path differ */
    if( !(r&(1<<3)) || r==(1<<3) )
      return be;
  }
  return NULL;
}

/* Quick delete ... */
static SC68browse_file_t *SC68browse_qdel(SC68browse_file_t *be, SC68browse_file_t *prev)
{
  if(be==NULL)
    return NULL;
  if(prev==NULL)
    bi.head = be->nxt;
  else
    prev->nxt = be->nxt;
  bi.n--;
  bi.natari -= !!(be->flags&(SC68BROWSE_YM|SC68BROWSE_MW));
  bi.namiga -= !!(be->flags&(SC68BROWSE_PL));
  bi.ntrack -= be->ntrack;
  bi.modified = 1;
  bi.time -= be->time;
  return be;
}

/* Del entry if from Browse-Info-Data-Base
 */
SC68browse_file_t *SC68browse_del(SC68browse_file_t *model)
{
  SC68browse_file_t *prev, *be;
  if(model==NULL) return NULL;
  for(prev=NULL, be=bi.head; be!=NULL && be!=model; (prev=be), (be=be->nxt) );
  return SC68browse_qdel(be,prev);
}

/* Add entry to Browse-Info-Data-Base
 */
SC68browse_file_t *SC68browse_add(SC68browse_file_t *be)
{
  if(be==NULL) return NULL;
/*  pdebug( "SC68browse_add((%03d)a:%s c:%s m:%s p:%s\n",
                 bi.n,be->aname,be->cname,be->mname,be->path);*/
  be->nxt = bi.head;
  bi.n++;
  bi.natari += !!(be->flags&(SC68BROWSE_YM|SC68BROWSE_MW));
  bi.namiga += !!(be->flags&(SC68BROWSE_PL));
  bi.ntrack += be->ntrack;
  bi.modified = 1;
  bi.time += be->time;
  return bi.head = be;
}

static char *next_space(char *s)
{
  for(;*s && !isspace(*s); s++);
  return s;
}

static char *kill_space(char *s)
{
  for(;isspace(*s); s++);
  return s;
}

static char *get_word(char *s)
{
  char *stop;
  s=kill_space(s);
  if(*s=='"')
  {
    s++;
    stop = strchr(s,'"');
    if(stop!=NULL) *stop=0;
    return s;
  }
  *next_space(s) = 0;
  return s;
}

/*
static int no_update(SC68browse_file_t *found, SC68browse_file_t *be)
{
  return !namecmp(found->path, be->path) &&
         !namecmp(found->aname, be->aname) &&
         !namecmp(found->mname, be->mname);
}
*/

static SC68browse_file_t *force_add(SC68browse_file_t *be)
{
  SC68browse_file_t *found;

  if( found=SC68browse_exist(be), found!=NULL)
  {
    int r;
    r = browse_cmp(be,found);
    /* Nothing change !!! */
    if(!r)
      return found;

    found = SC68browse_del(found);
    if(found==NULL)
    {
      SC68error_add("??? Internal : Can't delete existing ???\n");
      return NULL;
    }
    SC68_free(found);
  }

  if(found=new_entry(be), found==NULL)
    return NULL;

  return SC68browse_add(found);
}

/* read mm:sec time format
 * return time in sec or 0 for bad format
 */
static int read_time_format(char *a)
{
  char *deuxpoint,*s;
  int min=0,sec=0;
  if(deuxpoint=strchr(a,':'), deuxpoint==NULL)
    return 0;
  for(s=a; s<deuxpoint; s++)
    if(!isdigit(*s)) return -2;
    else min = min*10 + *s-'0';
  for(s++; isdigit(*s); s++)
    sec = sec*10 + *s-'0';
  if(s==deuxpoint)
    return 0;
  return min*60 + sec;
}

/* write mm:sec time format
 * return static char *
 */
static char *write_time_format(int time)
{
  static char s[8];
  if(time<=0)
    return strcpy(s,"--:--");
  else if(time>99*60+59)
    time=99*60+59;
  sprintf(s,"%02d:%02d",time/60,time%60);
  return s;
}

/* write big time format
 * return static char *
 */
static char *big_time_format(int time)
{
  return SC68utils_make_big_playing_time(time);
}

/* Read flags format form strring :
 * 'A' char means Amiga
 * 'Y' char means Yamaha 2149 (STF/E)
 * 'E' char means MicroWire (STE)
 * Function is not case sensitif.
 */
static int read_flags_format(char *a)
{
  int r=0, c;
  if(a==NULL) return SC68BROWSE_HWDEF;
  for(c=*a++; c; c=*a++)
  {
    c = toupper(c);
    if(c=='A') r |= SC68BROWSE_PL;
    if(c=='Y') r |= SC68BROWSE_YM;
    if(c=='E') r |= SC68BROWSE_MW;
  }
  return r;
}

/* write flags formnat string
 * see read_flags_format for more info
 * return static char *
 */
static char *write_flags_format(int flags)
{
  static char s[4];
  int i=0;
  if(flags&SC68BROWSE_PL) s[i++] = 'A';
  if(flags&SC68BROWSE_YM) s[i++] = 'Y';
  if(flags&SC68BROWSE_MW) s[i++] = 'E';
  s[i] = 0;
  return s;
}

/*
     returns:
        <0 : Error
         0 : EOF
         1 : OK
         2 : Empty line
*/
static int read_bi_line(FILE *f, SC68browse_file_t *be)
{
  char tmp[512],*s,*w;
  int  i,l;

  s=fgets(tmp,512,f);
  if(s==NULL) return 0;

  /* Clean */
  s[511]=0;
  for(i=0; i<512; i++)
    if(!isprint(s[i])) s[i]=0;
  if(strlen(s)==511)
    return SC68error_add("SC68browse_read: Line to long '%10s...'",s);
  s = kill_space(s);
  if(!strlen(s)) return 2;
  if(*s==';') return 2;

  /* Get author */
  w = get_word(s);
  l = strlen(w);
  if(!l)
    return SC68error_add("SC68browse_read: No author string");
  be->aname = w;
  s  = w+l+1;

  /* Get composer */
  w = get_word(s);
  l = strlen(w);
  if(!l)
    return SC68error_add("SC68browse_read: No composer string");
  /* String "-" , replique author name */
  if(strcmp(w,"-"))
    be->cname = w;
  else
    be->cname = be->aname;

  s  = w+l+1;

  /* Get music name */
  w = get_word(s);
  l = strlen(w);
  if(!l)
    return SC68error_add("SC68browse_read: No music string");
  be->mname=w;
  s  = w+l+1;

  /* Get file path */
  w = get_word(s);
  l = strlen(w);
  if(!l)
    return SC68error_add("SC68browse_read: No file path string");
  be->path=w;
  s  = w+l+1;

  /* Get playing time */
  w = get_word(s);
  l = strlen(w);
  if(!l)
    be->time = 0;
  else
    be->time=read_time_format(w);
  s  = w+l+1;

  /* Get number of tracks */
  w = get_word(s);
  l = strlen(w);
  if(!l)
    be->ntrack = 0;
  else
    be->ntrack=atoi(w);
  if((unsigned int)be->ntrack > 99)
    be->ntrack = 0;
  s  = w+l+1;

  /* Get HW-req flags */
  w = get_word(s);
  l = strlen(w);
  if(!l)
    be->flags = SC68BROWSE_HWDEF;
  else
    be->flags=read_flags_format(w);
  s  = w+l+1;

  if(force_add(be)==NULL)
    return -0x05;

  return 1;
}

static int browse_read(FILE *f)
{
  SC68browse_file_t  e;
  int err;
  do
  {
    err = read_bi_line(f,&e);
  } while(err>0);
  if(err) return -0x02;
  return 0;
}

/* SC68browse_read : Merge Browse info resource file
 */
int SC68browse_read(char *fname)
{
  int err;

  bi.f=SC68os_open_resource( SC68rsc_browseinfo,fname,"rt");

  /* Doesn't exist is not an error when default */
  if(bi.f==NULL)
  {
    if(fname!=NULL) return -0x01;
    SC68error_get(); /* Pop error */
    pdebug("SC68browse_read(NULL), No default\n");
    return 0;
  }

  err = browse_read(bi.f);
  fclose(bi.f);
  bi.f = NULL;
  return err;
}


static int browse_write(FILE *f)
{
  SC68browse_file_t  *be;
  int                nauthor=0;
  char               *prev;

  if(!bi.head)
    return 0;
  nauthor = 1;
  prev = bi.head->aname;
  for(be=bi.head->nxt; be!=NULL; be=be->nxt)
    if(namecmp(prev,be->aname))
    {
      nauthor++;
      prev = be->aname;
    }

  fprintf(  f,  "; SC68 Browse Info Data Base file - Generated by SC68\n"
                "; SC68 is a FREEWARE Mc68000 based computer Sound-Chip music emulator (Atari-ST & Amiga)\n"
                "; (c) Ben(jamin) Gerard / sashipa <ben@sashipa.com>\n;\n"
                "; - %d author%s in %d file%s (%d for Atari, %d for Amiga)\n"
                "; - %d track%s\n"
                "; - Playing time %s\n"
                ";\n",
                nauthor,(nauthor>1)?"s":"",
                bi.n,(bi.n>1)?"s":"",
                bi.natari,bi.namiga,
                bi.ntrack,(bi.ntrack>1)?"s":"",
                big_time_format(bi.time));

  for(be=bi.head; be!=NULL; be=be->nxt)
  {
    fprintf(  f,"\"%s\" \"%s\" \"%s\" \"%s\" \"%s\" \"%d\" \"%s\"\n",
           be->aname,
           be->cname==be->aname ? "-" : be->cname,
           be->mname,be->path,
           write_time_format(be->time),
           be->ntrack,
           write_flags_format(be->flags));
  }
  return 0;
}

/* SC68browse_write : Save all Browse info resource file
 */
int SC68browse_write(char *fname)
{
  /* Nothing !! Don't crash existing OR not modified */
  if(!bi.n || !bi.modified)
    return 0;

  if(SC68browse_sort( SC68browse_sort_by_author )<0)
    return -0x02;

  bi.f=SC68os_open_resource(SC68rsc_browseinfo,fname,"wt");
  if(bi.f==NULL)
    return -0x01;

  if(!browse_write(bi.f))
    bi.modified = 0;
  fclose(bi.f); bi.f=NULL;
  return 0;
}

/* SC68browse_write_file : hack for browse68 !!!
 */
int SC68browse_write_file(FILE *f)
{
  if(SC68browse_sort( SC68browse_sort_by_author )<0)
    return -0x02;
  if(f==NULL)
    return -0x03;
	return browse_write(f);
}



/* Qsort compatible fuction table, for parameters ordering
 */
static int (*sort_table[4])(
                         SC68browse_file_t **c1,
                         SC68browse_file_t **c2);

static volatile int level;

/* qsort compatible functions
 */
static int sort_by_author(SC68browse_file_t **c1, SC68browse_file_t **c2)
{
  return namecmp((*c1)->aname,(*c2)->aname);
}

static int sort_by_composer(SC68browse_file_t **c1, SC68browse_file_t **c2)
{
  return namecmp((*c1)->cname,(*c2)->cname);
}

static int sort_by_music(SC68browse_file_t **c1, SC68browse_file_t **c2)
{
  return namecmp((*c1)->mname,(*c2)->mname);
}

static int sort_by_path(SC68browse_file_t **c1, SC68browse_file_t **c2)
{
  return namecmp((*c1)->path,(*c2)->path);
}

static int sort_level(SC68browse_file_t **c1, SC68browse_file_t **c2)
{
  int r;
  for(level=0; level<4; level++)
      if(r=(sort_table[level])(c1,c2), r)
        return r;
  return 0;
}

/* Sort Browse info list :
 * order : -SC68browse_sort_by_author
 */
int SC68browse_sort(SC68browse_sort_t type)
{
  SC68browse_file_t **buf;
  SC68browse_file_t *be;
  int               i,nauthor;

  if(!bi.n)
    return 0;

  /* In order to count author, sort by author first !!! */
  if(type!=SC68browse_sort_by_author)
  {
    if(SC68browse_sort(SC68browse_sort_by_author)<0)
      return -0x10;
  }

  buf=(SC68browse_file_t **)SC68_malloc(bi.n*sizeof(SC68browse_file_t *));
  if(buf==NULL)
    return SC68error_add("SC68browse_sort(%d) : Malloc failed",type);

  /* Copy */
  for(i=0, be=bi.head; be!=NULL; i++, be=be->nxt)
    buf[i] = be;

  switch( type )
  {
    case SC68browse_sort_by_path:
      sort_table[0] = sort_by_path;
      sort_table[1] = sort_by_music;
      sort_table[2] = sort_by_author;
      sort_table[3] = sort_by_composer;
      break;

    case SC68browse_sort_by_music:
      sort_table[0] = sort_by_music;
      sort_table[1] = sort_by_path;
      sort_table[2] = sort_by_author;
      sort_table[3] = sort_by_composer;
      break;

    case SC68browse_sort_by_composer:
      sort_table[0] = sort_by_composer;
      sort_table[1] = sort_by_author;
      sort_table[2] = sort_by_music;
      sort_table[3] = sort_by_path;
      break;

    case SC68browse_sort_by_author:
    default:
      sort_table[0] = sort_by_author;
      sort_table[1] = sort_by_music;
      sort_table[2] = sort_by_path;
      sort_table[3] = sort_by_composer;
      break;
  }

  qsort(buf, bi.n, sizeof(SC68browse_file_t *),
        (int (*)(const void *, const void *))sort_level);

  /* Relink & Count author ( valid only if sort_by_author ) */
  nauthor = 1;
  bi.head = buf[0];
  for(i=0; i<bi.n-1; i++)
  {
    buf[i]->nxt = buf[i+1];
    nauthor += !!namecmp(buf[i]->aname,buf[i+1]->aname);
  }
  buf[i]->nxt = NULL;
  if(type==SC68browse_sort_by_author)
    bi.nauthor = nauthor;

  if(buf!=NULL) SC68_free(buf);

  return 0;
}

/* Add or Update entry into Browse Info Data Base
 */
int SC68browse_update(SC68browse_file_t *be)
/*char *aname, char *cname, char *mname, char *path, int time)*/
{
  if(be==NULL)
    return 0;
  if(be->cname==NULL)
  	be->cname = be->aname;

  if(force_add(be)==NULL)
  {
    SC68error_add("SC68browse_update(%s,%s) Failed",
         be->mname,be->path);
    return -1;
  }
  return 0;
}

/* Verify Browse Info Data Base integrity :
 * return # removed file...
 */
int SC68browse_verify(void)
{
  SC68browse_file_t *be,*prev,*nxt;
  int n=0;
  for(be=bi.head, prev=NULL; be!=NULL; be=nxt)
  {
    nxt = be->nxt;
    if(SC68file_verify(be->path)<0)
    {
      SC68_free(SC68browse_qdel(be,prev));
      n++;
    }
    else
      prev = be;
  }
  return n;
}

/* Generate HTML with current Browse Info
 * fname : Filename without extension
 */
int SC68browse_generate(char *fname)
{
  FILE *f;
  int  err;

  if(SC68browse_sort(SC68browse_sort_by_author)<0)
    return SC68error_add("SC68browse_generate() failed");

  f=SC68os_open_resource(SC68rsc_html,fname,"wt");

  if(f==NULL)
    return SC68error_add("SC68browse_generate() see previous error");

  err=SC68html_generate(f,&bi);

  SC68_fclose(f);
  return err;
}
