/**
 * @ingroup   file68_devel
 * @file      SC68utils.c
 * @author    Ben(jamin) Gerard <ben@sashipa.com>
 * @date      1999/03/08
 * @brief     sc68 miscellaneous utility functions
 * @version   $Id$
 */
 
/*
 *                 sc68 - Miscellaneous utility functions
 *         Copyright (C) 2001 Ben(jamin) Gerard <ben@sashipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */
 
#include <ctype.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "file68/SC68utils.h"
#include "file68/SC68error.h"

/*  Get system endianness
 */
int SC68utils_is_little_endian(void)
{
  int a = 1;
  return *(char *) &a;
}

int SC68utils_is_big_endian(void)
{
  return !SC68utils_is_little_endian();
}

/*  Compare two string without case : !!! No standard function for this !!!
 */
int SC68utils_strcmp(char *a, char *b)
{
  int av, bv;
  /* This is not very smart ... but what to do ? */
  if (a == b)
    return 0;
  if (a == NULL)
    return -1;
  if (b == NULL)
    return 1;
  for (av = toupper(*a), bv = toupper(*b); av && av == bv; a++, b++) {
    av = toupper(*a);
    bv = toupper(*b);
  }
  return av - bv;
}

/*  "a+b" with sizeof("a+b") <= l
 */
char *SC68utils_strcat(char *a, char *b, int l)
{
  int n;
  if (a == NULL || l < 0)
    return NULL;
  if (b == NULL)
    return a;
  for (n = strlen(a); n < l && *b; n++, b++)
    a[n] = *b;
  if (n < l)
    a[n] = 0;
  return a;
}


/* Construct a track-time info string
 * Format "TK MN:SS"
 * !!! This function return a static buffer don't crash it ( Hell with const ) !!!
 * track_num : 00 to 99] OR <0 to deactivate
 * seconds   : 00 to 5999 [99'59"] OR  <0 to deactivate
 */
char *SC68utils_make_track_time_info(int track_num, int seconds)
{
  static int old_track = -1, old_sec = -1;
  static char buffer[12] = "-- --:--";

  if (track_num < 0)
    track_num = -1;
  else if (track_num > 99)
    track_num = 99;
  if (track_num != old_track) {
    old_track = track_num;
    if (track_num < 0)
      buffer[0] = buffer[1] = '-';
    else {
      buffer[0] = '0' + track_num / 10;
      buffer[1] = '0' + track_num % 10;
    }
  }

  if (seconds < 0)
    seconds = -1;
  else if (seconds > 99 * 60 + 59)
    seconds = 99 * 60 + 59;
  if (seconds != old_sec) {
    old_sec = seconds;
    if (seconds < 0)
      strcpy(buffer + 3, "--:--");
    else
      sprintf(buffer + 3, "%02u:%02u", seconds / 60, seconds % 60);
  }

  buffer[8] = 0;
  return buffer;
}

/* Construct a playing-time info string from a tine in seconds
 * Format [D days, ][H h, ] MN' SS"
 *        0 => none
 * !!! This function return a static buffer don't crash it ( Hell with const ) !!!
 */
char *SC68utils_make_big_playing_time(int time)
{
  static char s[32];
  int sec, min, hr, day;
  if (time <= 0)
    return strcpy(s, "none");
  sec = time % 60;
  time /= 60;
  min = time % 60;
  time /= 60;
  hr = time % 24;
  time /= 24;
  day = time;
  if (day)
    sprintf(s, "%d day%s, %2dh, %02d' %02d\"", day, day > 1 ? "s" : "", hr,
	    min, sec);
  else if (hr)
    sprintf(s, "%2dh, %02d' %02d\"", hr, min, sec);
  else
    sprintf(s, "%02d' %02d\"", min, sec);
  return s;
}


/********************************************************
* MEMORY ALLOCATION
********************************************************/

/* malloc() */
void *SC68_malloc(int nb)
{
  void *b;
  if (nb <= 0) {
    SC68error_add("SC68_malloc(%d) : Weird malloc requested", nb);
    return NULL;
  }
  if (b = malloc(nb), b == NULL)
    SC68error_add("Memory allocation request (%d bytes) failed", nb);
  return b;
}

/* free() */
int SC68_free(void *b)
{
  if (b == NULL)
    return -1;
  free(b);
  return 0;
}

/********************************************************
* FILE OPERATION
********************************************************/

static char *errstr_fnull = "SC68_f%s : NULL pointer";
static char *errstr_fioerror = "SC68_f%s : %s error";
static char *errstr_toomanybytes =
    "SC68_f%s : Too many bytes requested (%d>%d)";

/* fopen() */
FILE *SC68_fopen(char *name, char *mode)
{
  static char *fstr = "open";
  FILE *f;
  if (f = fopen(name, mode), f == NULL) {
    SC68error_add(errstr_fioerror, fstr, name);
    return NULL;
  }
  return f;
}

/* fclose() */
int SC68_fclose(FILE * f)
{
  if (f != NULL)
    fclose(f);
  return 0;
}

/* fread() */
int SC68_fread(void *b, int nb, FILE * f)
{
  static char *fstr = "read";
  if (!nb)
    return 0;
  if (f == NULL || b == NULL)
    return SC68error_add(errstr_fnull, fstr);
  if (fread(b, nb, 1, f) != 1)
    return SC68error_add(errstr_fioerror, fstr, "Read");
  return 0;
}

/* fwrite() */
int SC68_fwrite(void *b, int nb, FILE * f)
{
  static char *fstr = "write";
  if (!nb)
    return 0;
  if (f == NULL || b == NULL)
    return SC68error_add(errstr_fnull, fstr);
  if (fwrite(b, nb, 1, f) != 1)
    return SC68error_add(errstr_fioerror, fstr, "Write");
  return 0;
}

/* flen() : size of file */
int SC68_flen(FILE * f)
{
  static char *fstr = "len";
  int pos, len;
  if (f == NULL)
    return SC68error_add(errstr_fnull, fstr);
  pos = ftell(f);
  fseek(f, 0L, SEEK_END);
  len = ftell(f);
  fseek(f, pos, SEEK_SET);
  return len;
}


/* Read from file a number of nbbytes bytes, in defined byte-order
 * nbbytes must not be greater than an int.
 */
int SC68_numberfread(int *pres, int nbbytes, int bigendian, FILE * f)
{
  static char *fstr = "numberread";
  int res = 0;
  unsigned char c;
  if (f == NULL)
    return SC68error_add(errstr_fnull, fstr);
  if (nbbytes > (int) sizeof(int))
    return SC68error_add(errstr_toomanybytes, fstr, nbbytes, sizeof(int));

  if (bigendian) {
    while (nbbytes--) {
      if (fread(&c, 1, 1, f) != 1)
	return SC68error_add(errstr_fioerror, fstr, "read");
      res = (res << 8) + c;
    }
  } else {
    int dec = (int) sizeof(int) * 8;
    while (nbbytes--) {
      if (fread(&c, 1, 1, f) != 1)
	return SC68error_add(errstr_fioerror, fstr, "read");
      res |= (int) c << (dec -= 8);
    }
    res >>= dec;
  }
  if (pres != NULL)
    *pres = res;
  return 0;
}

/* Write into file a number of nbbytes bytes, in defined byte-order
 * nbbytes must not be greater than an int.
 */
int SC68_numberfwrite(int val, int nbbytes, int bigendian, FILE * f)
{
  static char *fstr = "numberwrite";
  unsigned char c;
  if (f == NULL)
    return SC68error_add(errstr_fnull, fstr);
  if (nbbytes > (int) sizeof(int))
    return SC68error_add(errstr_toomanybytes, fstr, nbbytes, sizeof(int));

  if (bigendian) {
    int dec = nbbytes * 8;
    while (dec) {
      c = val >> (dec -= 8);
      if (fwrite(&c, 1, 1, f) != 1)
	goto error;
    }
  } else {
    while (nbbytes--) {
      c = val;
      val >>= 8;
      if (fwrite(&c, 1, 1, f) != 1)
	goto error;
    }
  }
  return 0;
error:
  return SC68error_add(errstr_fioerror, fstr, "write");
}
