/**
 * @ingroup   sc68app_devel
 * @file      SC68config.c
 * @author    Ben(jamin) Gerard<ben@sashipa.com>
 * @date      1999/07/27
 * @brief     sc68 - config file
 * @version   $Id$
 */

/*
 *                           sc68 - Config file
 *         Copyright (C) 2001 Ben(jamin) Gerard <ben@sashipa.com>
 *
 *  This program is free software; you can redistribute it and/or modify it
 *  under the terms of the GNU General Public License as published by the
 *  Free Software Foundation; either version 2 of the License, or (at your
 *  option) any later version.
 *
 *  This program is distributed in the hope that it will be useful, but
 *  WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 *  General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License along
 *  with this program; if not, write to the Free Software Foundation, Inc.,
 *  59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 *
 */

#include <stdlib.h>
#include <ctype.h>
#include <string.h>

#include "sc68app/SC68.h"
#include "sc68app/SC68config.h"
#include "file68/SC68error.h"
#include "file68/SC68utils.h"


/* define for the config default value */

#define SAMPLING_RATE   44100
#define BIT_PER_SAMPLE  1       /* 0:8 1:16 */
#define CHANNELS        2
#define BUFFER          4096
#define MIX_PASS        50
#define AMIGA_BLEND     (2<<16)/3

#define TRACK_LOOP      2       /* 0:none 1:stop 2:next-track 3:restart */
#define DISK_LOOP       2       /* 0:none 1:stop 2:next-disk 3:restart */

#define PLAY_AT_LOAD    1
#define	DEFAULT_TIME	(2*60)
#define FORCE_TRACK     0       /* 0:no forced track */

#define TIME_MODE       1 /* 0:elapsed 1:remaining */
#define SKIP_TIME	4 /* Skip music time in sec */

#define DELAY_BUFFER     4096
#define DELAY_TIME	 6
#define DELAY_STRENGTH   164
#define DELAY_PROCESSING 0
#define DELAY_LR         0

#define BROWSE_INFO      0

#define FFT_SIZE         7
#define FFT_MOTION       (int)(0.12f*256.0f)
#define FFT_LCOLOR       -1
#define FFT_RCOLOR       -1
#define PCM_LCOLOR       -1
#define PCM_RCOLOR       -1


static SC68config_entry_t tab[] =
{
  /* version number */
  {"version","major*100+minor", 0, 10000, 0},

  /* digital audio config */
  {"sampling_rate","sample output rate in HZ", 8000, 48000, SAMPLING_RATE},
  {"bit_per_sample","bit/sample 0:8 bit 1:16 bit", 0, 1, BIT_PER_SAMPLE},
  {"channels","# channels 1:mono 2:stereo", 1, 2, CHANNELS },
  {"buffer","buffer size in sample", 1024, 65536, BUFFER },
  {"mix_pass","maximun mix pass before polling", 1, 64, MIX_PASS },
  {"amiga_blend","left/right voices blending factor for Amiga",0,65536,AMIGA_BLEND },

  /* Play interface config */
  {"force_track", "override default track (0:OFF)", 0, 99/*SC68_MAX_TRACK*/, FORCE_TRACK },
  {"play_at_load", "auto-play track when disk is loaded", 0, 1, PLAY_AT_LOAD },
  {"skip_time","don't auto-play music which time in seconds is less than", 0, SC68_MAX_TIME },
  {"eot_action","action at end of track 0:none 1:stop 2:next-track 3:restart", 0, 3, TRACK_LOOP },
  {"eod_action","action at end of disk 0:none 1:stop 2:next-disk 3:restart",0 , 3, DISK_LOOP },

  /* Time interface config */
  {"time_mode","time display mode 0:elapsed 1:remaining", 0, 1, TIME_MODE },
  {"default_time","default time in seconds for unknown", 0, SC68_MAX_TIME, DEFAULT_TIME },

  /* Delay processing */
  {"delay_buffer", "size of delay buffer in sample (32 bit)", 2048, 16384, DELAY_BUFFER },
  {"delay_time","delay time in mix buffer length factor",-32, 32, DELAY_TIME},
  {"delay_strength","delay strength factor", 64, 192, DELAY_STRENGTH},
  {"delay_processing","delay processing at startup 1:on 0:off", 0, 1, DELAY_PROCESSING},
  {"delay_lr","left to right delay spacializer 0:on 1:off", 0, 1, DELAY_LR},

  /* Browse info */
  {"browse_info", "Generate browse data base 1:ON 0:0FF", 0, 1, BROWSE_INFO},

  /* Funky FX */
  {"fft.size","FFT precision in power of 2", 4, 10, FFT_SIZE},
  {"fft.motion","FFT motion speed", 0, 256, FFT_MOTION},
  {"fft.lcolor","left RGB color of FFT display", -1, 0xFFFFFF, FFT_LCOLOR},
  {"fft.rcolor","right RGB color of FFT display", -1, 0xFFFFFF, FFT_RCOLOR},
  {"pcm.lcolor","Left RGB color of PCM display", -1, 0xFFFFFF, PCM_LCOLOR},
  {"pcm.rcolor","right RGB color of PCM display", -1, 0xFFFFFF, PCM_RCOLOR},

  {"total_time","total of music played since first launch", 0, 0, 0 },

};

static const int nconfig = sizeof(tab)/sizeof(tab[0]);

static char previous_err[] = "%s : See Previous";
static int config_check(SC68config_t *conf)
{
  if(conf==NULL)
    return SC68error_add("SC68config_xxx() : NULL pointer");
  return 0;
}

/* Check config values and correct invalid ones
 */
int SC68config_valid(SC68config_t *conf)
{
  static char fct_name[] = "SC68config_valid";
  int err,i;
  if(err=config_check(conf), err<0)
    return SC68error_add(previous_err, fct_name);
  for(i=0; i<nconfig; i++)
  {
    if(tab[i].min==0 && tab[i].max==0)
      continue;
    else if(tab[i].min==0 && tab[i].max==1) /* Boolean */
      (*conf)[i] = !! (*conf)[i];
    else if((*conf)[i]<tab[i].min)
      (*conf)[i] = tab[i].min;
    else if((*conf)[i]>tab[i].max)
      (*conf)[i] = tab[i].max;
  }
  return 0;
}

/* Get indice of named field in SC68config_t array
*  return <0 if field doesn't exist
 */
int SC68config_get_id(char *name)
{
  int i;
  if(name!=NULL)
    for(i=0; i<nconfig; i++)
      if(!SC68utils_strcmp(name,tab[i].name)) return i;
  return SC68error_add("SC68config_get_id(%s) Not found", name==NULL ? "<null>" : name);
}

/* Output config to FILE
 */
int SC68config_write(FILE *f, SC68config_t *conf)
{
  int i,err;
  static char fct_name[] = "SC68config_write";
  if(err=SC68config_valid(conf), err<0)
    return SC68error_add(previous_err, fct_name);
  for(i=0; i<nconfig; i++)
    {
      int idf;
      char *format[3] =
      {
	"%-20s=%6d   [%5d..%5d] (%5d) %s\n",
	"%-20s= 0x%06x [%3d..%x]    (%2d) %s\n",
	"%-20s=%6d   [%4d..%x]    (%2d) %s\n"
      };
      if(tab[i].max <= 0x10000) idf = 0;
      else if ((*conf)[i]>=0)  idf = 1;
      else                     idf = 2;
      fprintf(f,format[idf],
	      tab[i].name,(*conf)[i],tab[i].min,tab[i].max,tab[i].def,tab[i].comment);
    }
  return 0;
}

/* Output config to stdout
 */
int SC68config_display(SC68config_t *conf)
{
  return SC68config_write(stdout,conf);
}


/* Load config from file
 */
int SC68config_load(SC68config_t *conf, char *name)
{
  static char fct_name[] = "SC68config_load";
  FILE *f;
  char s2[1024];
  int err;

  if(err=SC68config_default(conf), err<0)
    return SC68error_add(previous_err, fct_name);

  if(f=SC68_fopen(name,"rt"), f==NULL)
    return SC68error_add(previous_err, fct_name);

  err = 0;
  while(!err && !feof(f))
  {
    char *s;
    *s2 = 0;
    if(s=fgets(s2,sizeof(s2),f), (s!=NULL && *s))
    {
      while(isspace(*s)) s++;
      if(*s!='#')
      {
        int v;
        char *s2;
        for(s2=s; isgraph(*s2) && *s2 != '=' ; s2++);
        if(*s2!='=')
        {
          *s2++=0;
          s2=strchr(s2,'=');
        }
        if(s2!=NULL && *s2!=0)
        {
          int i;
          *s2++ = 0;
	  v = strtol(s2,&s2,0);
          if(i=SC68config_get_id(s), i>=0)
            (*conf)[i] = v;
        }
      }
    }
  }
  if(!err) err = SC68config_valid(conf);
  SC68_fclose(f);
  return err;
}

/* Save config into file
 */
int SC68config_save(SC68config_t *conf, char *name)
{
  static char fct_name[] = "SC68config_save";
  FILE *f;
  int err;

  if(err=SC68config_valid(conf), err<0)
    return SC68error_add(previous_err, fct_name);

  if(f=SC68_fopen(name,"wt"), f==NULL)
    return SC68error_add(previous_err, fct_name);
  fprintf(f,
    "# SC68 config file.\n"
    "# (c)Ben / sashipa <ben@sashipa.com>\n"
    "#\n"
    "# http://sashipa.ben.free.fr\n"
    "#\n"
    "# You can edit this file. If you remove it, sc68 will generate\n"
    "# a new one at start-up with default values, but you will lost your\n"
    "# total playing time. To avoid it, you should either save it value\n"
    "# or delete all lines you want to be resetted. You can use 'C' number\n"
    "# format (e.g.0xFE0120 in color parameters).\n"
    "#\n");
  SC68config_write(f,conf);
  fprintf(f,"# End of SC68 config file\n#\n");
  SC68_fclose(f);
  return 0;
}

/* Fill config struct with default value
 */
int SC68config_default(SC68config_t *conf)
{
  static char fct_name[] = "SC68config_default";
  int err,i;
  if(err=config_check(conf), err<0)
    return SC68error_add(previous_err, fct_name);
  for(i=0; i<nconfig; i++)
    (*conf)[i] = tab[i].def;
  return SC68config_valid(conf);
}
